"""NDG XACML ElementTree based Rule Element reader 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: rulereader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.rule import Rule, Effect
from ndg.xacml.core.condition import Condition
from ndg.xacml.core.target import Target
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader
from ndg.xacml.parsers.etree.factory import ReaderFactory


class RuleReader(ETreeAbstractReader):
    '''ElementTree based XACML Rule parser
    
    @cvar TYPE: XACML type to instantiate from parsed object
    @type TYPE: type
    '''
    TYPE = Rule
    
    def __call__(self, obj):
        """Parse rule object
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML expression instance
        @rtype: ndg.xacml.core.rule.Rule derived type 
        @raise XMLParseError: error reading element               
        """
        elem = super(RuleReader, self)._parse(obj)
        
        xacmlType = RuleReader.TYPE
        rule = xacmlType()
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
        
        # Unpack *required* attributes from top-level element
        attributeValues = []
        for attributeName in (xacmlType.RULE_ID_ATTRIB_NAME, 
                              xacmlType.EFFECT_ATTRIB_NAME):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLParseError('No "%s" attribute found in "%s" '
                                        'element' %
                                        (attributeName,
                                         xacmlType.ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue) 
        
        rule.effect = Effect()        
        rule.id, rule.effect.value = attributeValues
            
        # Parse sub-elements
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            
            if localName == xacmlType.DESCRIPTION_LOCAL_NAME:
                if childElem.text is not None:
                    rule.description = childElem.text.strip()
                    
            elif localName == Condition.ELEMENT_LOCAL_NAME:
                ConditionReader = ReaderFactory.getReader(Condition)
                rule.condition = ConditionReader.parse(childElem)
                                   
            elif localName == Target.ELEMENT_LOCAL_NAME:
                TargetReader = ReaderFactory.getReader(Target)
                rule.target = TargetReader.parse(childElem)
            
            else:
                raise XMLParseError("XACML Rule child element name %r not "
                                    "recognised" % localName)                
                
        return rule


"""NDG XACML ElementTree based reader for ActionAttributeDesignator type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "19/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: actionattributedesignatorreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.attributedesignator import ActionAttributeDesignator
from ndg.xacml.parsers.etree.attributedesignatorreader import \
    AttributeDesignatorReaderBase


class ActionAttributeDesignatorReader(AttributeDesignatorReaderBase): 
    '''ElementTree based XACML Action Attribute Designator type parser
    
    @cvar TYPE: XACML class type that this reader will read values into
    @type TYPE: abc.ABCMeta
    '''
    TYPE = ActionAttributeDesignator


"""NDG XACML one and only functions module

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "01/04/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: round.py 7088 2010-06-25 12:14:03Z pjkersha $'
from ndg.xacml.core.functions import (AbstractFunction, 
                                      FunctionClassFactoryInterface)
from ndg.xacml.core.context.exceptions import XacmlContextTypeError


class Round(AbstractFunction):
    """Base class for XACML <type>-round functions
    
    @cvar FUNCTION_NS: namespace for this function
    @type FUNCTION_NS: string
    """
    FUNCTION_NS = AbstractFunction.V1_0_FUNCTION_NS + 'round'
    
    def evaluate(self, num):
        """Check a bag has one element only and return it
        
        @param num: number to round up
        @type num: int / long / float
        @rtype: float
        @raise TypeError: incorrect type for input
        """
        try:
            return round(num)
        except TypeError, e:
            raise XacmlContextTypeError('Round function: %s' % e)

    
class FunctionClassFactory(FunctionClassFactoryInterface):
    """Class Factory for round XACML function class
    """
    def __call__(self, identifier):
        '''Create class for the Round XACML function identifier
        
        @param identifier: XACML round function identifier
        @type identifier: basestring
        @return: round function class or None if identifier doesn't match
        @rtype: ndg.xacml.core.functions.v1.round.Round / NoneType
        '''
        if identifier == Round.FUNCTION_NS:
            return Round
        else:
            return None



"""NDG XACML ElementTree based reader for AttributeDesignator type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: attributedesignatorreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.utils import str2Bool
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.expressionreader import ExpressionReader


class AttributeDesignatorReaderBase(ExpressionReader):
    '''ElementTree based XACML Attribute Designator base class type parser
    '''
    def _parseExtension(self, elem, attributeDesignator):
        """Parse Attribute Designator element
        
        @param elem: ElementTree XML element
        @type elem: xml.etree.Element
        
        @param attributeDesignator: attribute designator
        @type attributeDesignator: ndg.xacml.core.attributedesignator.AttributeDesignator
        
        @raise XMLParseError: error parsing attribute ID XML attribute
        
        @return: updated attribute designator
        @rtype: ndg.xacml.core.attributedesignator.AttributeDesignator
        """
        xacmlType = self.__class__.TYPE
        
        # Unpack additional *required* attributes from top-level element
        attributeValues = []
        for attributeName in (xacmlType.ATTRIBUTE_ID_ATTRIB_NAME,):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLParseError('No "%s" attribute found in "%s" element' %
                                    (attributeName, 
                                     xacmlType.ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue)
             
        attributeDesignator.attributeId, = attributeValues
              
        # Optional attributes
        issuer = elem.attrib.get(xacmlType.ISSUER_ATTRIB_NAME)
        if issuer is not None:
            attributeDesignator.issuer = issuer
             
        mustBePresent = elem.attrib.get(xacmlType.MUST_BE_PRESENT_ATTRIB_NAME)
        if mustBePresent is not None:
            attributeDesignator.mustBePresent = str2Bool(mustBePresent)
   
        return attributeDesignator




"""NDG XACML 2.0 regular expression matching function module

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "26/03/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: regexp_match.py 7088 2010-06-25 12:14:03Z pjkersha $'
from ndg.xacml.core.functions import FunctionClassFactoryBase

# Use v1.0 schema base class for version 2.0 additional types
from ndg.xacml.core.functions.v1.regexp_match import RegexpMatchBase
    

class FunctionClassFactory(FunctionClassFactoryBase):
    """Class Factory for *-regexp-match XACML function classes
    
    @cvar FUNCTION_NAMES: regular expression match function URNs
    @type FUNCTION_NAMES: tuple
    
    @cvar FUNCTION_NS_SUFFIX: generic suffix for one and only function URNs
    @type FUNCTION_NS_SUFFIX: string
    
    @cvar FUNCTION_BASE_CLASS: base class for regular expression match function 
    classes 
    @type FUNCTION_BASE_CLASS: ndg.xacml.core.functions.v1.RegexMatchBase
    """
    FUNCTION_NS_SUFFIX = RegexpMatchBase.FUNCTION_NS_SUFFIX
    FUNCTION_BASE_CLASS = RegexpMatchBase
    FUNCTION_NAMES = (
        'urn:oasis:names:tc:xacml:2.0:function:anyURI-regexp-match',
        'urn:oasis:names:tc:xacml:2.0:function:ipAddress-regexp-match',
        'urn:oasis:names:tc:xacml:2.0:function:dnsName-regexp-match',
        'urn:oasis:names:tc:xacml:2.0:function:rfc822Name-regexp-match',
        'urn:oasis:names:tc:xacml:2.0:function:x500Name-regexp-match'
    )


#!/usr/bin/env python
"""NDG XACML Policy unit test package 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: test_policy.py 7087 2010-06-25 11:23:09Z pjkersha $"
import unittest
from os import path
import logging
logging.basicConfig(level=logging.DEBUG)

from ndg.xacml.core.policy import Policy
from ndg.xacml.core.attributedesignator import SubjectAttributeDesignator
from ndg.xacml.core.attributeselector import AttributeSelector
from ndg.xacml.parsers.etree.factory import ReaderFactory

from ndg.xacml.test import XACML_NDGTEST1_FILEPATH, THIS_DIR


class XACMLPolicyTestCase(unittest.TestCase):
    XACML_TEST1_FILENAME = "rule1.xml"
    XACML_TEST1_FILEPATH = path.join(THIS_DIR, XACML_TEST1_FILENAME)
    XACML_TEST2_FILENAME = "rule2.xml"
    XACML_TEST2_FILEPATH = path.join(THIS_DIR, XACML_TEST2_FILENAME)
    XACML_TEST3_FILENAME = "rule3.xml"
    XACML_TEST3_FILEPATH = path.join(THIS_DIR, XACML_TEST3_FILENAME)
    XACML_TEST4_FILENAME = "rule4.xml"
    XACML_TEST4_FILEPATH = path.join(THIS_DIR, XACML_TEST4_FILENAME)
   
    def test01ETreeParseRule1Policy(self):
        PolicyReader = ReaderFactory.getReader(Policy)
        policy = PolicyReader.parse(XACMLPolicyTestCase.XACML_TEST1_FILEPATH)
        self.assert_(policy)
        
        self.assert_(
            policy.policyId == "urn:oasis:names:tc:example:SimplePolicy1")
        
        self.assert_(policy.ruleCombiningAlgId == \
        "urn:oasis:names:tc:xacml:1.0:rule-combining-algorithm:deny-overrides")
        
        self.assert_(
            "Med Example Corp access control policy" in policy.description)
        
        self.assert_(len(policy.target.subjects) == 0)
        
        self.assert_(policy.rules[0].id == \
                     "urn:oasis:names:tc:xacml:2.0:example:SimpleRule1")
        
        self.assert_(policy.rules[0].effect == 'Permit')
        
        self.assert_(
            'Any subject with an e-mail name in the med.example.com domain' in \
            policy.rules[0].description)
        
        self.assert_(len(policy.rules[0].target.subjects) == 1)
        self.assert_(len(policy.rules[0].target.actions) == 0)
        self.assert_(len(policy.rules[0].target.resources) == 0)
        self.assert_(len(policy.rules[0].target.environments) == 0)
        
        self.assert_(len(policy.rules[0].target.subjects[0
                                                         ].subjectMatches) == 1)
        
        self.assert_(policy.rules[0].target.subjects[0].subjectMatches[0
            ].matchId == \
            "urn:oasis:names:tc:xacml:1.0:function:rfc822Name-match")
        
        self.assert_(policy.rules[0].target.subjects[0].subjectMatches[0
            ].attributeValue.dataType == \
            "urn:oasis:names:tc:xacml:1.0:data-type:rfc822Name")
        
        self.assert_(policy.rules[0].target.subjects[0].subjectMatches[0
            ].attributeDesignator.dataType == \
            "urn:oasis:names:tc:xacml:1.0:data-type:rfc822Name")
        
        # Attribute ID
        self.assert_(policy.rules[0].target.subjects[0].subjectMatches[0
            ].attributeDesignator.attributeId == \
            "urn:oasis:names:tc:xacml:1.0:subject:subject-id")
         
    def test02ETreeParseRule2Policy(self):
        PolicyReader = ReaderFactory.getReader(Policy)
        policy = PolicyReader.parse(XACMLPolicyTestCase.XACML_TEST2_FILEPATH)
        self.assert_(policy)
        
        self.assert_(
        policy.policyId == "urn:oasis:names:tc:xacml:2.0:example:policyid:2")
        
        self.assert_(policy.ruleCombiningAlgId == \
        "urn:oasis:names:tc:xacml:1.0:rule-combining-algorithm:deny-overrides")
        
        self.assert_(policy.description is None)
        
        self.assert_(len(policy.target.actions) == 0)
        
        self.assert_(policy.rules[0].id == \
                     "urn:oasis:names:tc:xacml:2.0:example:ruleid:2")
        
        self.assert_(policy.rules[0].effect == 'Permit')
        
        self.assert_(policy.rules[0].description == """\
A person may read any medical record in the
            http://www.med.example.com/records.xsd namespace
            for which he or she is the designated parent or guardian, 
            and for which the patient is under 16 years of age""")
        
        self.assert_(len(policy.rules[0].target.subjects) == 0)
        self.assert_(len(policy.rules[0].target.actions) == 1)
        self.assert_(len(policy.rules[0].target.resources) == 1)
        self.assert_(len(policy.rules[0].target.environments) == 0)
        
        self.assert_(len(policy.rules[0].target.resources[0
                                                    ].resourceMatches) == 2)
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[0
            ].matchId == "urn:oasis:names:tc:xacml:1.0:function:string-equal")
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[0
            ].attributeValue.dataType == \
                                    "http://www.w3.org/2001/XMLSchema#string")
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[0
            ].attributeValue.value == 'urn:med:example:schemas:record')
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[0
            ].attributeDesignator.dataType == \
                                    "http://www.w3.org/2001/XMLSchema#string")
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[1
            ].attributeDesignator.attributeId == \
                            "urn:oasis:names:tc:xacml:1.0:resource:xpath")
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[1
            ].matchId == \
                "urn:oasis:names:tc:xacml:1.0:function:xpath-node-match")
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[1
            ].attributeValue.dataType == \
                                    "http://www.w3.org/2001/XMLSchema#string")
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[1
            ].attributeValue.value == '/md:record')
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[1
            ].attributeDesignator.dataType == \
                                    "http://www.w3.org/2001/XMLSchema#string")
        
        self.assert_(policy.rules[0].target.resources[0].resourceMatches[1
            ].attributeDesignator.attributeId == \
                                "urn:oasis:names:tc:xacml:1.0:resource:xpath")
        
        # Verify Action
        self.assert_(len(policy.rules[0].target.actions[0
                                                    ].actionMatches) == 1)
        
        self.assert_(policy.rules[0].target.actions[0].actionMatches[0
            ].matchId == "urn:oasis:names:tc:xacml:1.0:function:string-equal")
        
        self.assert_(policy.rules[0].target.actions[0].actionMatches[0
            ].attributeValue.dataType == \
                                    "http://www.w3.org/2001/XMLSchema#string")
        
        self.assert_(policy.rules[0].target.actions[0].actionMatches[0
            ].attributeValue.value == "read")
        
        self.assert_(policy.rules[0].target.actions[0].actionMatches[0
            ].attributeDesignator.dataType == \
                                    "http://www.w3.org/2001/XMLSchema#string")
        
        self.assert_(policy.rules[0].target.actions[0].actionMatches[0
            ].attributeDesignator.attributeId == \
                            "urn:oasis:names:tc:xacml:1.0:action:action-id")

        self.assert_(policy.rules[0].condition)        
        self.assert_(policy.rules[0].condition.expression.functionId == \
                     "urn:oasis:names:tc:xacml:1.0:function:and")
        
        self.assert_(len(policy.rules[0].condition.expression.expressions) == 1)
        
        self.assert_(policy.rules[0].condition.expression.expressions[0
            ].functionId == \
                'urn:oasis:names:tc:xacml:1.0:function:string-equal')
        
        self.assert_(len(policy.rules[0].condition.expression.expressions) == 1)
        
        self.assert_(len(policy.rules[0].condition.expression.expressions[0
                     ].expressions) == 2)
        
        self.assert_(policy.rules[0].condition.expression.expressions[0
            ].expressions[0].functionId == \
                "urn:oasis:names:tc:xacml:1.0:function:string-one-and-only")
        
        self.assert_(isinstance(
                        policy.rules[0].condition.expression.expressions[0
                            ].expressions[0
                            ].expressions[0], SubjectAttributeDesignator))
        
        self.assert_(policy.rules[0].condition.expression.expressions[0
                            ].expressions[0
                            ].expressions[0].attributeId == \
                            "urn:oasis:names:tc:xacml:2.0:example:attribute:"
                            "parent-guardian-id")

        self.assert_(policy.rules[0].condition.expression.expressions[0
                            ].expressions[0
                            ].expressions[0].dataType == \
                            "http://www.w3.org/2001/XMLSchema#string")
        
        self.assert_(policy.rules[0].condition.expression.expressions[0
                            ].expressions[0
                            ].expressions[0].attributeId == \
                            "urn:oasis:names:tc:xacml:2.0:example:attribute:"
                            "parent-guardian-id")
        
        self.assert_(isinstance(policy.rules[0
                            ].condition.expression.expressions[0
                            ].expressions[1
                            ].expressions[0], AttributeSelector))
       
        self.assert_(policy.rules[0
                            ].condition.expression.expressions[0
                            ].expressions[1
                            ].expressions[0].requestContextPath == \
                            "//md:record/md:parentGuardian/md:parentGuardianId/"
                            "text()")
        
        self.assert_(policy.rules[0
                            ].condition.expression.expressions[0
                            ].expressions[1
                            ].expressions[0].dataType == \
                            "http://www.w3.org/2001/XMLSchema#string")

    def test03ETreeParseRule3Policy(self):
        PolicyReader = ReaderFactory.getReader(Policy)
        
        try:
            policy = PolicyReader.parse(XACMLPolicyTestCase.XACML_TEST3_FILEPATH)
            self.assert_(policy)
        except NotImplementedError, e:
            print("Expecting Obligations not implemented exception: %s" %e)
                    
    def test04ETreeParseRule4Policy(self):
        PolicyReader = ReaderFactory.getReader(Policy)
        policy = PolicyReader.parse(XACMLPolicyTestCase.XACML_TEST4_FILEPATH)
        self.assert_(policy)
                    
    def test05ETreeParseNdg1Policy(self):
        # Example policy for URI Regular expression based matching of
        # resources for NDG
        PolicyReader = ReaderFactory.getReader(Policy)
        policy = PolicyReader.parse(XACML_NDGTEST1_FILEPATH)
        self.assert_(policy)    

        
if __name__ == "__main__":
    unittest.main()


"""NDG XACML context package defines classes for types in the access control
context schema

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7087 2010-06-25 11:23:09Z pjkersha $"
from ndg.xacml.utils import TypedList
from ndg.xacml.core import XacmlCoreBase
from ndg.xacml.core.attribute import Attribute


class XacmlContextBase(XacmlCoreBase):
    """Base class for XACML Request and Response types
    
    @cvar ELEMENT_LOCAL_NAME: XML local element name, derived classes should
    set
    @type ELEMENT_LOCAL_NAME: None"""
    ELEMENT_LOCAL_NAME = None
    __slots__ = ()
    
    def __init__(self):
        """ELEMENT_LOCAL_NAME check makes this class virtual - derived classes
        must override this method and set ELEMENT_LOCAL_NAME to the appropriate
        string
        """
        if self.__class__.ELEMENT_LOCAL_NAME is None:
            raise NotImplementedError('Set "ELEMENT_LOCAL_NAME" in a derived '
                                      'type')
    
   
class RequestChildBase(XacmlContextBase):
    """Base class for XACML Context Subject, Resource, Action and Environment
    types
    
    @ivar __attributes: XACML Context subject attributes
    @type __attributes: ndg.xacml.utils.TypedList
    """
    __slots__ = ('__attributes', )
    
    def __init__(self):
        """Initialise attribute list"""
        self.__attributes = TypedList(Attribute)
        
    @property
    def attributes(self):
        """
        @return: XACML Context subject attributes
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__attributes
    